import inspect
import sys

from prefect import task


def to_prefect_tasks(module_name, general_module):
    this_module = sys.modules[module_name]
    general_funcs = inspect.getmembers(general_module, inspect.isfunction)
    for name, func in general_funcs:
        docstring = getattr(func, "__doc__", "")
        first_line = docstring.splitlines()[0]
        task_func = task(func, name=first_line)
        setattr(this_module, name, task_func)



import os
from tempfile import TemporaryDirectory

import pandas as pd
import synapseclient

CONTENT_TYPES = {",": "text/csv", "\t": "text/tab-separated-values"}


def bundle_client_args(auth_token, **kwargs):
    """Synapse - Bundle client arguments"""
    client_args = dict(authToken=auth_token, silent=True)
    client_args.update(kwargs)
    return client_args


def get_dataframe(client_args, synapse_id, sep=None):
    """Synapse - Download and load data frame"""
    client = synapseclient.login(**client_args)
    file = client.get(synapse_id, downloadFile=False)
    file_handle_id = file._file_handle.id
    file_temp_url = client.restGET(
        f"/file/{file_handle_id}",
        client.fileHandleEndpoint,
        params={
            "fileAssociateId": synapse_id,
            "fileAssociateType": "FileEntity",
            "redirect": False,
        },
    )
    data_frame = pd.read_table(file_temp_url, sep=sep)
    return data_frame


def store_dataframe(client_args, data_frame, name, parent_id, sep=","):
    """Synapse - Serialize and upload data frame"""
    client = synapseclient.login(**client_args)
    with TemporaryDirectory() as dirname:
        fpath = os.path.join(dirname, name)
        content_type = CONTENT_TYPES.get(sep, None)
        with open(fpath, "w") as f:
            data_frame.to_csv(f, sep=sep, index=False)
        syn_file = synapseclient.File(
            fpath, name=name, parent=parent_id, contentType=content_type
        )
        syn_file = client.store(syn_file)
        os.remove(fpath)
    return syn_file



from sagetasks.sevenbridges.utils import SbgUtils


def bundle_client_args(auth_token, platform="cavatica", endpoint=None, **kwargs):
    """SevenBridges - Bundle client arguments"""
    return SbgUtils.bundle_client_args(auth_token, platform, endpoint, **kwargs)


def get_project_id(client_args, project_name, billing_group_name):
    """SevenBridges - Get (or create) project"""
    utils = SbgUtils(client_args)
    project = utils.get_or_create_project(project_name, billing_group_name)
    project_id = utils.extract_id(project)
    return project_id


def get_copied_app_id(client_args, project, app_id=None):
    """SevenBridges - Get (or import) an imported copy of a public app"""
    utils = SbgUtils(client_args)
    utils.open_project(project)
    copied_app = utils.get_or_create_copied_app(app_id)
    copied_app_id = utils.extract_id(copied_app)
    return copied_app_id


def get_volume_id(client_args, volume_name=None, volume_id=None):
    """SevenBridges - Get a cloud volume"""
    utils = SbgUtils(client_args)
    volumes = utils.get_volume(volume_name, volume_id)
    assert len(volumes) > 0, "This function cannot create a volume if it's missing"
    assert len(volumes) < 2, "Use a more specific volume name or use an ID instead"
    volume = volumes[0]
    volume_id = utils.extract_id(volume)
    return volume_id


def import_volume_file(client_args, project, volume_id, volume_path, project_path):
    """SevenBridges - Import a file from a volume"""
    utils = SbgUtils(client_args)
    utils.open_project(project)
    imported_file = utils.get_or_create_volume_file(
        volume_id, volume_path, project_path
    )
    imported_file_id = utils.extract_id(imported_file)
    return imported_file_id


def create_tasks(client_args, project, app_id, manifest, inputs_fn):
    """SevenBridges - Create draft tasks"""
    utils = SbgUtils(client_args)
    utils.open_project(project)
    draft_task_ids = list()
    for task_name, inputs, callback_fn in inputs_fn(utils.client, manifest):
        task = utils.get_or_create_task(app_id, inputs, task_name, callback_fn)
        task_id = utils.extract_id(task)
        draft_task_ids.append(task_id)
    return draft_task_ids



# --------------------------------------------------------------
# Imports
# --------------------------------------------------------------

# basic imports
from sys import argv

import pandas as pd
from prefect import Flow, Parameter, task, unmapped
from prefect.tasks.secrets import EnvVarSecret

import sagetasks.sevenbridges.prefect as sbg

# specific task function imports
import sagetasks.synapse.prefect as syn
from sagetasks.sevenbridges.inputs import manifest_to_kf_rnaseq_app_inputs_factory

# --------------------------------------------------------------
# Generate custom functions using factories
# --------------------------------------------------------------


prepare_task_inputs = manifest_to_kf_rnaseq_app_inputs_factory()


# --------------------------------------------------------------
# Define custom task functions
# --------------------------------------------------------------


@task
def print_columns(data_frame):
    print(data_frame.columns.tolist())


@task
def head_df(data_frame, n):
    return data_frame.head(n)


@task
def split_rows(data_frame):
    return [row for _, row in data_frame.iterrows()]


@task
def print_values(x):
    print(f"\n{x}\n")


@task
def prepare_file_imports(row):
    s3_uri_prefix = "s3://include-sandbox/synapse/"
    row["volume_path"] = row.s3_uri.replace(s3_uri_prefix, "", 1)
    row["project_path"] = "synapse/" + row.component + "/" + row.filepath
    return row


@task
def call_import_volume_file(sbg_args, project_id, volume_id, row):
    imported_file_id = sbg.import_volume_file.run(
        sbg_args,
        project_id,
        volume_id,
        row.volume_path,
        row.project_path,
    )
    row["cavatica_file_id"] = imported_file_id
    return row


@task
def concat_rows(rows):
    return pd.concat(rows, axis=1).T


# --------------------------------------------------------------
# Open a Flow context and use the functional API (if possible)
# --------------------------------------------------------------

with Flow("Demo") as flow:

    # Parameters
    manifest_id = Parameter("manifest_id")
    synapse_folder = Parameter("synapse_folder")
    project_name = Parameter("project_name")
    billing_group_name = Parameter("billing_group_name")
    app_id = Parameter("app_id")
    volume_name = Parameter("volume_name")

    # Secrets
    syn_token = EnvVarSecret("SYNAPSE_AUTH_TOKEN")
    sbg_token = EnvVarSecret("SB_AUTH_TOKEN")

    # Client arguments
    syn_args = syn.bundle_client_args(syn_token)
    sbg_args = sbg.bundle_client_args(sbg_token)

    # Extract
    manifest = syn.get_dataframe(syn_args, manifest_id, sep=",")
    project_id = sbg.get_project_id(sbg_args, project_name, billing_group_name)
    app_id = sbg.get_copied_app_id(sbg_args, project_id, app_id)
    volume_id = sbg.get_volume_id(sbg_args, volume_name)

    # Transform
    rows = split_rows(manifest)
    rows_prep = prepare_file_imports.map(rows)

    # Load
    rows_imp = call_import_volume_file.map(
        unmapped(sbg_args),
        unmapped(project_id),
        unmapped(volume_id),
        rows_prep,
    )
    sbg_manifest = concat_rows(rows_imp)
    drafted_tasks = sbg.create_tasks(
        sbg_args, project_id, app_id, sbg_manifest, prepare_task_inputs
    )
    print_values(drafted_tasks)
    syn.store_dataframe(syn_args, sbg_manifest, "sbg_manifest.csv", synapse_folder)

params = {
    "manifest_id": "syn33335513",
    "synapse_folder": "syn33335225",
    "project_name": "include-sandbox",
    "billing_group_name": "include-dev",
    "app_id": "cavatica/apps-publisher/kfdrc-rnaseq-workflow",
    "volume_name": "include_sandbox_ro",
}

if len(argv) < 2 or argv[1] == "run":
    flow.run(parameters=params)
elif argv[1] == "viz":
    flow.visualize(filename="flow", format="png")
elif argv[1] == "register":
    flow.register(project_name="demo")
else:
    print("Available Modes: 'run', 'viz', and 'register'")



import re
from copy import deepcopy

import numpy as np

# Parameters used in GTEx pipeline for normal (i.e., non-tumor) samples
GTEX_NORMAL_PARAMS = {
    "alignInsertionFlush": "None",
    "alignIntronMax": 1000000,
    "alignMatesGapMax": 1000000,
    "alignSJDBoverhangMin": 1,
    "alignSJoverhangMin": 8,
    "alignSoftClipAtReferenceEnds": "Yes",
    "chimJunctionOverhangMin": 15,
    "chimMainSegmentMultNmax": 1,
    "chimSegmentMin": 15,
    "limitSjdbInsertNsj": 1200000,
    "outFilterIntronMotifs": "None",
    "outFilterMatchNminOverLread": 0.33,
    "outFilterMismatchNmax": 999,
    "outFilterMismatchNoverLmax": 0.1,
    "outFilterMultimapNmax": 20,
    "outFilterScoreMinOverLread": 0.33,
    "outFilterType": "BySJout",
    "twopassMode": "Basic",
}

STRANDEDNESS_DEFAULTS = ("default", "rf-stranded", "fr-stranded")

TUMOR_REGEX = re.compile(r"tumou?r")


def get_unique_value(df, col, default=None):
    """Retrieves unique scalar value from Pandas DataFrame column."""
    if col not in df and default is not None:
        raw = [default]
    else:
        raw = df[col].unique()
    assert len(raw) == 1
    val = raw[0]
    if isinstance(val, np.int64):
        val = int(val)
    return val


def format_rg_val(val):
    """Replaces all whitespace from input string with underscores."""
    return re.sub(r"\s", "_", val)


# TODO: Currently assumes paired-end data
# TODO: Separate function to import suggested reference files
# TODO: Add support for (and validate) different versions of the workflow
# TODO: Add mapping between app IDs (and versions?) to factory functions
def manifest_to_kf_rnaseq_app_inputs_factory(
    file_col="cavatica_file_id",
    sample_col="sample_id",
    readlen_col="read_length",
    sampletype_col="sample_type",
    orientation_col="read_orientation",
    orientation_vals=("R1", "R2"),
    strandedness_col="strandedness",
    strandedness_vals=STRANDEDNESS_DEFAULTS,
    library_col="library_id",
    platform_col="platform",
):
    """Creates function for generating the KF RNA-seq app inputs from a file manifest.

    Args:
        file_col (str, optional): Manifest column name for Cavatica file IDs.
            Defaults to "cavatica_file_id".
        sample_col (str, optional): Manifest column name for sample IDs.
            Defaults to "sample_id".
        readlen_col (str, optional): Manifest column name for read lengths.
            Defaults to "read_length".
        sampletype_col (str, optional): Manifest column name for sample types
            (e.g., tumor vs normal). Defaults to "sample_type".
        orientation_col (str, optional): Manifest column name for read orientations.
            Defaults to "read_orientation".
        orientation_vals (tuple, optional): Manifest column values for mapping to
            R1 and R2. Defaults to ("R1", "R2").
        strandedness_col (str, optional): Manifest column name for RNA-seq strandedness.
            Defaults to "strandedness".
        strandedness_vals (list of str, optional): Manifest column values for mapping
            to the defaults. Defaults to ("default", "rf-stranded", "fr-stranded").
        library_col (str, optional): Manifest column name for library IDs.
            Defaults to "library_id".
        platform_col (str, optional): Manifest column name for sequencing platforms.
            Defaults to "platform".
    """

    def manifest_to_kf_rnaseq_app_inputs(client, manifest):
        """Prepares KF RNA-seq app inputs from file manifest."""
        # Unpacking valid values
        r1_val, r2_val = orientation_vals
        strandedness_map = dict(zip(strandedness_vals, STRANDEDNESS_DEFAULTS))

        # Prepare inputs
        by_sample = manifest.groupby(sample_col)
        for sample_id in by_sample.groups:
            # Subset manifest for specific sample
            sample_df = by_sample.get_group(sample_id)
            assert len(sample_df.index) == 2
            # Retrieve Cavatica file IDs
            r1_file_id = sample_df[sample_df[orientation_col] == r1_val][file_col].iat[
                0
            ]
            r2_file_id = sample_df[sample_df[orientation_col] == r2_val][file_col].iat[
                0
            ]
            # Retrieve other metadata
            strandedness_raw = get_unique_value(sample_df, strandedness_col)
            strandedness = strandedness_map.get(strandedness_raw, None)
            assert strandedness
            library_id = get_unique_value(sample_df, library_col, "Not_Reported")
            platform = get_unique_value(sample_df, platform_col, "Not_Reported")
            read_length = get_unique_value(sample_df, readlen_col)
            sample_type = get_unique_value(sample_df, sampletype_col)
            is_tumor = TUMOR_REGEX.search(sample_type)
            # Prepare params dictionary
            inputs = dict() if is_tumor else deepcopy(GTEX_NORMAL_PARAMS)
            updates = {
                "input_type": "FASTQ",
                "reads1": client.files.get(r1_file_id),
                "reads2": client.files.get(r2_file_id),
                "runThreadN": 36,
                "wf_strand_param": strandedness,
                "sample_name": sample_id,
                "rmats_read_length": read_length,
                "outSAMattrRGline": "\t".join(
                    map(
                        format_rg_val,
                        [
                            f"ID:{sample_id}",
                            f"LB:{library_id}",
                            f"PL:{platform}",
                            f"SM:{sample_id}",
                        ],
                    )
                ),
            }
            inputs.update(updates)
            task_name = "kf-rnaseq-workflow - " + sample_id

            def callback_fn(task):
                task.inputs["output_basename"] = task.id
                task.save()

            yield task_name, inputs, callback_fn

    return manifest_to_kf_rnaseq_app_inputs



import re
import time
from functools import partial
from pathlib import PurePosixPath

import sevenbridges as sbg
from sevenbridges import ImportExportState
from sevenbridges.http.error_handlers import maintenance_sleeper, rate_limit_sleeper
from sevenbridges.meta.transformer import Transform
from sevenbridges.models.project import Project

ENDPOINTS = {
    "cavatica": "https://cavatica-api.sbgenomics.com/v2",
    "cgc": "https://cavatica-api.sbgenomics.com/v2",
    "sevenbridges": "https://api.sbgenomics.com/v2",
}


class SbgUtils:
    def __init__(self, client_args) -> None:
        """Initializes the SevenBridges client with the bundled information.

        `client_args` can be generated with the `bundle_client_args()` static method.

        Optionally, you can set a default project for various methods with the
        `open_project()` method.
        """
        self.client = sbg.Api(
            **client_args, error_handlers=[rate_limit_sleeper, maintenance_sleeper]
        )
        self._project = None

    def extract_id(self, resource):
        """Extracts the resource ID (or returns the ID if already a string).

        These IDs are strings, which can be pickled unlike the resource objects.
        """
        resource_id = Transform.to_resource(resource)
        return resource_id

    def get_or_create(self, get_fn, create_fn):
        """Gets a single resource or creates it if missing."""
        collection = get_fn()
        if len(collection) == 0:
            create_fn()
            collection = get_fn()
            assert len(collection) == 1
            result = collection[0]
        elif len(collection) == 1:
            result = collection[0]
        else:
            raise ValueError("There shouldn't be more than one match.")
        return result

    @staticmethod
    def bundle_client_args(auth_token, platform="cavatica", endpoint=None, **kwargs):
        """Bundles the information for authenticating a SevenBridges client."""
        assert platform is None or platform in ENDPOINTS
        assert platform or endpoint
        endpoint = endpoint if endpoint else ENDPOINTS[platform]
        client_args = dict(url=endpoint, token=auth_token, **kwargs)
        return client_args

    def get_billing_group(self, billing_group_name):
        """Retrieves the billing groups with the given name."""
        billing_groups = self.client.billing_groups.query()
        matches = [bg for bg in billing_groups if bg.name == billing_group_name]
        return matches

    def _get_project_by_id(self, project_id):
        """Retrieves the project with the given ID."""
        try:
            project = self.client.projects.get(project_id)
            projects = project
        except sbg.errors.NotFound:
            projects = []
        return projects

    def _get_project_by_name(self, project_name):
        """Retrieves the projects with the given name."""
        projects = self.client.projects.query(name=project_name)
        return projects

    def get_project(self, project_name=None, project_id=None):
        """Retrieves the projects with the given name or ID."""
        assert project_name or project_id  # At least one is given
        assert not (project_name and project_id)  # Both cannot be given
        if project_name:
            matches = self._get_project_by_name(project_name)
        else:
            matches = self._get_project_by_id(project_id)
        return matches

    def create_project(self, project_name, billing_group_name):
        """Creates a project with the given name and billing group."""
        billing_group = self.get_billing_group(billing_group_name)
        project = self.client.projects.create(project_name, billing_group)
        return project

    def get_or_create_project(self, project_name, billing_group_name):
        """Gets (or creates) a project with the given information."""
        get_fn = partial(self.get_project, project_name)
        create_fn = partial(self.create_project, project_name, billing_group_name)
        return self.get_or_create(get_fn, create_fn)

    @property
    def project(self):
        """Property for raising an error if a project hasn't been opened yet."""
        if self._project is None:
            raise ValueError("Project not opened yet. Use `open_project()`.")
        return self._project

    def open_project(self, project):
        """Sets the given project as the default for other methods."""
        project_id = self.extract_id(project)
        project = self.get_project(project_id=project_id)
        self._project = project

    def get_public_app(self, app_id):
        """Retrieves the public app with the given ID."""
        public_apps = self.client.apps.query(visibility="public", id=app_id)
        assert len(public_apps) == 1
        public_app = public_apps[0]
        return public_app

    def _get_apps_by_name(self, app_name):
        """Retrieves the private apps with the given name."""
        project_apps = self.client.apps.query(project=self.project, q=app_name)
        return project_apps

    def _get_app_suffix(self, app_slug, increment=False):
        """Generates an incrementing suffix for public apps that are copied."""
        apps = self._get_apps_by_name(app_slug)
        regex = re.compile(app_slug + r"-(\d)+")
        matches = [regex.search(x.id) for x in apps]
        matches = [x for x in matches if x]
        if matches:
            versions = [int(x[1]) for x in matches]
            versions.sort()
            last_version = versions[-1]
            if increment:
                last_version += 1
            suffix = f"-{last_version}"
        else:
            suffix = ""
            if increment:
                suffix = "-1"
        return suffix

    def get_copied_app_name(self, app_id, increment=False):
        """Generates a consistent naming scheme for public apps that are copied."""
        public_app = self.get_public_app(app_id)
        app_slug = public_app.id.split("/")[-1]
        suffix = self._get_app_suffix(app_slug, increment)
        app_name = f"{app_slug}{suffix}"
        return app_name

    def get_copied_app(self, app_id):
        """Retrieves a public app that's been copied to a project."""
        app_name = self.get_copied_app_name(app_id)
        apps = self.client.apps.query(project=self.project, q=app_name)
        # If multiple projects exist, pick the one with the shortest name
        apps = [x for x in apps if not x.raw.get("sbg:archived", False)]
        if len(apps) > 1:
            apps = sorted(apps, key=lambda x: len(x.id))
            apps = apps[:1]
        return apps

    def import_app(self, app_id):
        """Copies a public app into the opened project."""
        public_app = self.get_public_app(app_id)
        app_name = self.get_copied_app_name(app_id, increment=True)
        project_app = public_app.copy(project=self.project, name=app_name)
        return project_app

    def get_or_create_copied_app(self, app_id):
        """Gets (or copies) a public app in the opened project."""
        get_fn = partial(self.get_copied_app, app_id)
        create_fn = partial(self.import_app, app_id)
        return self.get_or_create(get_fn, create_fn)

    def _get_volume_by_id(self, volume_id):
        """Retrieves the cloud volume with the given ID."""
        try:
            volume = self.client.volumes.get(volume_id)
            volumes = [volume]
        except sbg.errors.NotFound:
            volumes = []
        return volumes

    def _get_volume_by_name(self, volume_name):
        """Retrieves the cloud volumes with the given name."""
        volumes = self.client.volumes.query()
        volumes = [vol for vol in volumes if vol.name == volume_name]
        return volumes

    def get_volume(self, volume_name=None, volume_id=None):
        """Retrieves the cloud volumes with the given name or ID."""
        assert volume_name or volume_id  # At least one is given
        assert not (volume_name and volume_id)  # Both cannot be given
        if volume_name:
            matches = self._get_volume_by_name(volume_name)
        else:
            matches = self._get_volume_by_id(volume_id)
        return matches

    def _get_parent_args(self, parent):
        """Generates relevant function kwargs for projects or folders."""
        if isinstance(parent, Project):
            parent_args = {"project": parent}
        else:
            parent_args = {"parent": parent}
        return parent_args

    def get_folder(self, folder_name, parent):
        """Retrieves the folder with the given name and parent."""
        parent_args = self._get_parent_args(parent)
        children = self.client.files.query(**parent_args)
        folders = [x for x in children if getattr(x, "type", None) == "folder"]
        matches = [x for x in folders if x.name == folder_name]
        return matches

    def create_folder(self, folder_name, parent):
        """Creates a folder with the given name and parent."""
        parent_args = self._get_parent_args(parent)
        folder = self.client.files.create_folder(name=folder_name, **parent_args)
        return folder

    def get_or_create_folder(self, folder_name, parent):
        """Gets (or creates) a folder with the given name and parent."""
        get_fn = partial(self.get_folder, folder_name, parent)
        create_fn = partial(self.create_folder, folder_name, parent)
        return self.get_or_create(get_fn, create_fn)

    def get_folders_recursively(self, folder_names, parent=None):
        """Gets (or creates) a nested list of folders recursively."""
        if parent is None:
            parent = self.project
        for folder_name in folder_names:
            folder = self.get_or_create_folder(folder_name, parent)
            parent = folder
        return folder

    def get_file(self, file_name, parent):
        """Retrieves a file with the given name and parent."""
        parent_args = self._get_parent_args(parent)
        children = self.client.files.query(**parent_args)
        files = [x for x in children if getattr(x, "type", None) == "file"]
        matches = [x for x in files if x.name == file_name]
        return matches

    def _wait_for_import_job(self, import_job):
        """Waits for the volume file import job to complete (successfully or not)."""
        while True:
            import_state = import_job.reload().state
            if import_state in (ImportExportState.COMPLETED, ImportExportState.FAILED):
                break
            time.sleep(5)
        return import_job

    def _get_imported_file(self, import_job):
        """Retrieves the imported file with the given import job."""
        if import_job.state == ImportExportState.COMPLETED:
            imported_file = import_job.result
        elif import_job.state == ImportExportState.FAILED:
            raise sbg.SbgError("Failed to import file from volume")
        else:
            raise sbg.SbgError("Import job not complete yet")
        return imported_file

    def import_volume_file(self, volume_id, volume_path, parent):
        """Imports the given volume file under the given parent."""
        import_job = self.client.imports.submit_import(
            volume=volume_id, parent=parent, location=volume_path
        )
        import_job = self._wait_for_import_job(import_job)
        imported_file = self._get_imported_file(import_job)
        return imported_file

    def get_or_create_volume_file(self, volume_id, volume_path, project_path):
        """Gets (or imports) the given volume file under the given project path."""
        project_path = PurePosixPath(project_path)
        dir_name = project_path.parent
        file_name = project_path.name
        folder_names = dir_name.parts
        parent = self.get_folders_recursively(folder_names)
        get_fn = partial(self.get_file, file_name, parent)
        create_fn = partial(self.import_volume_file, volume_id, volume_path, parent)
        return self.get_or_create(get_fn, create_fn)

    def get_task(self, task_name=None, app_id=None):
        """Retrieves the tasks with the given name and/or app ID."""
        matches = self.client.tasks.query(project=self.project)
        if task_name:
            matches = [t for t in matches if t.name == task_name]
        if app_id:
            matches = [t for t in matches if app_id in t.app]
        return matches

    def create_task(self, app_id, inputs, task_name, callback_fn=None):
        """Drafts a task with the given app, inputs, and task name.

        The optional `callback_fn` provides the option to update the task
        once it's created (e.g., updating an input using the task ID).
        """
        task = self.client.tasks.create(
            name=task_name, project=self.project, app=app_id, inputs=inputs, run=False
        )
        if callback_fn:
            callback_fn(task)
        return task

    def get_or_create_task(self, app_id, inputs, task_name, callback_fn=None):
        """Gets (or drafts) a task with the given app, inputs, and task name."""
        get_fn = partial(self.get_task, task_name, app_id)
        create_fn = partial(self.create_task, app_id, inputs, task_name, callback_fn)
        return self.get_or_create(get_fn, create_fn)
